// Server.cpp (Test Server)

#include "StdAfx.h"

#include "Server.h"

using namespace System::Runtime::CompilerServices;

using namespace TestServer;

// ClientMirrorImage methods

ClientMirrorImage::ClientMirrorImage(int nInstanceNumber, Timers::Timer ^tmrHeartbeat) :
  m_nInstanceNumber(nInstanceNumber), m_tmrHeartbeat(tmrHeartbeat), m_nTimeout(60000), m_status(ClientMirrorStatus::Active)
{
  tmrHeartbeat->Elapsed += gcnew Timers::ElapsedEventHandler(this, &ClientMirrorImage::HeartbeatHandler);
}

[MethodImpl(MethodImplOptions::Synchronized)]
String ^ClientMirrorImage::ToString()
{
  String ^str = String::Format("Client #{0}\t{1}", m_nInstanceNumber, m_status);
  if (m_status == ClientMirrorStatus::Active)
    str += String::Format(", timeout {0} seconds", m_nTimeout / 1000);
  return str;
}

[MethodImpl(MethodImplOptions::Synchronized)]
void ClientMirrorImage::HeartbeatHandler(Object ^sender, Timers::ElapsedEventArgs ^e)
{
  if (m_status == ClientMirrorStatus::Active && m_nTimeout > 0) {
    m_nTimeout -= m_tmrHeartbeat->Interval;
    if (m_nTimeout <= 0) {
      m_status = ClientMirrorStatus::TimedOut;
      DetachFromHeartbeat();
    }
  }
}

void ClientMirrorImage::DetachFromHeartbeat()
{
  if (m_tmrHeartbeat) {
    m_tmrHeartbeat->Elapsed -= gcnew Timers::ElapsedEventHandler(this, &ClientMirrorImage::HeartbeatHandler);
    m_tmrHeartbeat = nullptr;
  }
}

// Asserts that the object is active and optionally resets the timeout

[MethodImpl(MethodImplOptions::Synchronized)]
void ClientMirrorImage::AssertActive(bool bResetTimeout)
{
  if (!IsActive())
    throw gcnew InvalidOperationException(String::Format("Attempted to access inactive ClientMirrorImage object (#{0})",
      m_nInstanceNumber));
  if (bResetTimeout)
    m_nTimeout = 60000;
}

void ClientMirrorImage::Detach()
{
  DetachFromHeartbeat();
  if (IsActive())
    m_status = ClientMirrorStatus::Detached;
}

bool ClientMirrorImage::IsActive()
{
  return m_status == ClientMirrorStatus::Active;
}

// Server methods

Server::Server() : m_nClientInstance(0), m_strSharedData(String::Empty),
  m_lstClientRegistry(gcnew List<ClientMirrorImage ^>), m_tmrHeartbeat(gcnew Timers::Timer(1000)), m_nTimeout(60000)
{
  if (!s_svrPrimaryServerInstance) s_svrPrimaryServerInstance = this;
  Interlocked::Increment(s_nServerObjectCount);
  m_tmrHeartbeat->Elapsed += gcnew Timers::ElapsedEventHandler(this, &Server::HeartbeatHandler);
  m_tmrHeartbeat->Start();
}

Server::~Server()
{
  Interlocked::Decrement(s_nServerObjectCount);
}

array<String ^> ^Server::GetInstanceInfo()
{
  return s_svrPrimaryServerInstance ? s_svrPrimaryServerInstance->GetInstanceInfoCore() :
    gcnew array<String ^>{"No server object instantiated yet"};
}

[MethodImpl(MethodImplOptions::Synchronized)]
array<String ^> ^Server::GetInstanceInfoCore()
{
  List<String ^> ^lst = gcnew List<String ^>;
  String ^str = "Server\t";
  if (IsIdle()) {
    if (m_nTimeout > 0)
      str += String::Format("Idle, timeout {0} seconds", m_nTimeout / 1000);
    else {
      str += "TimedOut";
      if (s_dltShutdownRequestCallback) str += ", shutting down...";
    }
  } else str += "Active";
  lst->Add(str);
  for each (ClientMirrorImage ^cmi in m_lstClientRegistry)
    lst->Add(cmi->ToString());
  return lst->ToArray();
}

bool Server::IsIdle()
{
  if (!m_lstClientRegistry->Count) return true;
  int i = 0;
  while (i < m_lstClientRegistry->Count && !m_lstClientRegistry[i]->IsActive()) ++i;
  return i >= m_lstClientRegistry->Count;
}

bool Server::IsPrimaryServerInstanceIdle()
{
  return !s_svrPrimaryServerInstance || s_svrPrimaryServerInstance->IsIdle();
}

[MethodImpl(MethodImplOptions::Synchronized)]
void Server::HeartbeatHandler(Object ^sender, Timers::ElapsedEventArgs ^e)
{
  if (IsIdle()) {
    if (m_nTimeout > 0) {
      if ((m_nTimeout -= m_tmrHeartbeat->Interval) <= 0 && s_dltShutdownRequestCallback)
        s_dltShutdownRequestCallback->BeginInvoke(nullptr, nullptr);
    }
  } else m_nTimeout = 60000;
}

int Server::GetServerObjectCount()
{
  return s_nServerObjectCount;
}

// Service contract interface method implementations

int Server::GetClientInstanceNumber()
{
  m_lstClientRegistry->Add(gcnew ClientMirrorImage(++m_nClientInstance, m_tmrHeartbeat));
  if (s_dltStatusReportCallback) s_dltStatusReportCallback(Thread::CurrentThread->ManagedThreadId);
  return m_nClientInstance;
}

String ^Server::GetSharedData(int nInstanceNumber)
{
  if (s_dltStatusReportCallback) s_dltStatusReportCallback(Thread::CurrentThread->ManagedThreadId);
  m_lstClientRegistry[nInstanceNumber - 1]->AssertActive(true);
  return m_strSharedData;
}

void Server::SetSharedData(int nInstanceNumber, String ^strData)
{
  if (s_dltStatusReportCallback) s_dltStatusReportCallback(Thread::CurrentThread->ManagedThreadId);
  m_lstClientRegistry[nInstanceNumber - 1]->AssertActive(true);
  m_strSharedData = strData;
}

void Server::Detach(int nInstanceNumber)
{
  m_lstClientRegistry[nInstanceNumber - 1]->Detach();
}

// To be called upon start-up of the supposedly singular primary server .exe instance,
// returns true if no other instance is running yet, i.e. the "launched" event was not set:

bool Server::Startup(Action ^dltShutdownRequestCallback, Action<int> ^dltStatusReportCallback)
{
  EventWaitHandle ^evt = gcnew EventWaitHandle(false, EventResetMode::ManualReset, "WcfShareTestServerLaunchedEvent");
  if (evt->WaitOne(0)) {
    evt->Close();
    return false;  // Start-up failed: primary instance already running
  }
  s_evtServerLaunched = evt;
  s_dltShutdownRequestCallback = dltShutdownRequestCallback;
  s_dltStatusReportCallback = dltStatusReportCallback;
  s_host = gcnew ServiceHost(Server::typeid, gcnew Uri("net.pipe://localhost/TestServer"));
  s_host->Open();
  evt->Set();
  return true;
}

// To be called upon shutdown of the singular server .exe instance:

void Server::Shutdown()
{
  if (s_evtServerLaunched) {
    s_evtServerLaunched->Reset();
    s_evtServerLaunched->Close();
  }
  if (s_host) s_host->Close();
}
